'''Various utility functions for handling Debian-like archive structures.

Copyright (C) 2006 Canonical Ltd.
Author: Martin Pitt <martin.pitt@ubuntu.com>

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2 of the License, or (at your
option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
the full text of the license.
'''

import apt_pkg
import urllib
import os
import gzip
import bz2
import lzma
import tempfile
import os.path
import shutil
import tarfile
import time
import subprocess
import glob
import logging
from io import StringIO

apt_pkg.init_system()


def debcontrol_map(url, map_key='Package'):
    '''Read a debcontrol format file from the given URL and parse it into a map
    map_key->key->value.

    map_key defines a 'primary key' of the possible keys in the file.
    E. g. debcontrol_map('Sources.gz')['dpkg']['Version'] would deliver the
    Version of the dpkg source package.
    '''
    # write the uncompressed downloaded list into a temporary file
    if url.startswith('http://') or url.startswith('ftp://') or \
            url.startswith('https://') or url.startswith('file://'):
        data = tempfile.NamedTemporaryFile()
        src = urllib.request.urlopen(url, timeout=60)
        data.write(src.read())
        src.close()
        data.flush()
        data.seek(0)
    else:
        data = open(url, mode='rb')

    tagfile = tempfile.TemporaryFile()
    if url.endswith('.gz'):
        tagfile.write(gzip.GzipFile(fileobj=data, mode='rb').read())
    elif url.endswith('.bz2'):
        tagfile.write(bz2.BZ2File(data.name).read())
    elif url.endswith('.xz'):
        tagfile.write(lzma.LZMAFile(data.name, mode='rb').read())
    else:
        tagfile.write(data.read())
    tagfile.seek(0)
    data.close()

    map = {}
    parser = apt_pkg.TagFile(tagfile)
    while parser.step():
        pkg = {}
        for key in parser.section.keys():
            if key != map_key:
                pkg[key] = parser.section.get(key)
        if pkg:
            map[parser.section.get(map_key)] = pkg
    tagfile.close()
    return map


def archive_map(series_info, mirrors=None, pockets_filter=None,
                strip_epoch=False, ignore_missing=False):
    '''Build map of supported package versions

    @series_info: seriesname -> (components, architectures) map
    @mirrors: list of mirrors to check; defaults to
              standard Ubuntu mirrors
    @strip_epoch: If True, remove epochs from versions.
    @ignore_missing: If True, ignore missing Packages files.

    Return packagename -> pocket -> version -> arch -> component map.
    '''
    pockets = ['', '-updates', '-proposed']

    if mirrors is None:
        mirrors = ['http://archive.ubuntu.com/ubuntu',
                   'http://ports.ubuntu.com']

    archive_map = {}
    for series, (components, arches) in series_info.items():
        for p in pockets:
            if pockets_filter is not None and (series + p) not in pockets_filter:
                logging.debug('Ignoring pocket %s%s, not in filter', series, p)
                continue
            for c in components:
                for a in arches:
                    for m in mirrors:
                        url = '%s/dists/%s%s/%s/binary-%s/Packages' % (
                            m, series, p, c, a)
                        try:
                            index = debcontrol_map(url + '.xz')
                            logging.debug('archive_map: Got index %s.xz', url)
                            break
                        except IOError as e:
                            # fall back to gz compressed
                            try:
                                index = debcontrol_map(url + '.gz')
                                logging.debug('archive_map: Got index %s.gz', url)
                                break
                            except IOError as e:
                                # fall back to no compression
                                try:
                                    index = debcontrol_map(url)
                                    logging.debug('archive_map: Got index %s', url)
                                    break
                                except IOError as e:
                                    logging.debug('archive_map: index %s failed: %s', url, e)
                    else:
                        if not ignore_missing:
                            logging.error('cannot find dists/%s%s/%s/binary-%s/Packages on any mirror',
                                          series, p, c, a)
                        continue

                    for pkg, info in index.items():
                        try:
                            ver = info['Version']
                        except KeyError:
                            logging.warning('package %s in %s has no Version: field', pkg, url)
                            continue
                        if strip_epoch and ':' in ver:
                            ver = ver.split(':', 1)[1]
                        archive_map.setdefault(pkg, {}).setdefault(series + p, {}).setdefault(ver, {})[a] = c

    return archive_map


def debcontrol_dominate(infile, outfile, group_by='Package',
                        sort_key='Version', cmp=apt_pkg.version_compare):
    '''Read a file-like object in debcontrol syntax, group the records by the
    key group_by and only retain the member with the largest sort_key value
    per group.

    The result is written to 'outfile' (which must be a file-like object). The
    comparison defaults to apt_pkg.version_compare (since this should be the
    main use case, but another signature-compatible comparison function may be
    specified.
    '''
    max = {}  # group_by -> (sort_key_value, text)
    parser = apt_pkg.TagFile(infile)
    while parser.step():
        if cmp(max.get(parser.section[group_by], ('', ''))[0], parser.section[sort_key]) < 0:
            max[parser.section[group_by]] = (parser.section[sort_key], str(parser.section))

    ks = sorted(max)
    first = True
    for k in ks:
        if first:
            first = False
        else:
            outfile.write('\n')
        outfile.write(max[k][1])


def install_pool(archive_root, file_path, component, source):
    '''Install given file into the pool, with the given component and source
    package name.

    Return True on success, or False if the target file already exists.
    '''
    if source.startswith('lib'):
        hash = source[:4]
    else:
        hash = source[0]

    destdir = os.path.join(archive_root, 'pool', component, hash, source)
    destfile = os.path.join(destdir, os.path.basename(file_path))
    if os.path.exists(destfile):
        return False
    if not os.path.isdir(destdir):
        os.makedirs(destdir)

    with open(file_path, 'rb') as infile, open(destfile, 'wb') as outfile:
        outfile.write(infile.read())

    return True


def create_indexes(archive_root, pocket, components, archs, ext='.deb',
                   stderr=subprocess.PIPE, ref_archive=None, gpg_sign=False,
                   gpg_keyid=None):
    '''Create an apt-ftparchive configuration file, create the necessary
    directory structures and all apt-ftparchive to create Packages.gz files.

    ext specifies the file name extension (defaulting to '.deb').
    stderr allows you to redirect the apt-ftparchive output (e. g.
    None to use the caller's stderr instead of ignoring it).

    If ref_archive is given (result from archive_map() without epochs),
    create_indexes() will only consider the package versions in that map.

    If gpg_sign is True, "gpg -abs --batch -o Release.gpg Release" will be
    called on Release files. If gpg_keyid is specified, it is passed as
    --default-key.
    '''
    if ref_archive and 'source' in archs:
        raise NotImplementedError('create_index() does not support source packages with a reference_archive')

    # create configuration file
    conf = tempfile.NamedTemporaryFile(mode='w+')
    print('''Dir {
        ArchiveDir ".";
        CacheDir ".cache";
        FileListDir "./lists";
}

Default {
        Packages::Extensions "%s";
        Packages::Compress ". gzip xz";
        Sources::Compress ". gzip xz";
}

TreeDefault {
        Directory "./pool";
        SrcDirectory "./pool";
        BinCacheDB "db";
        Contents " ";
        FileList "$(SECTION)-$(ARCH)";
}

Tree "dists/%s" {
        Sections "%s";
        Architectures "%s";
}''' % (ext, pocket, ' '.join(components), ' '.join(archs)), file=conf)
    conf.flush()

    old_cwd = os.getcwd()
    os.chdir(archive_root)
    try:
        # create directories and file lists
        if not os.path.isdir('lists'):
            os.mkdir('lists')
        if not os.path.isdir('.cache'):
            os.mkdir('.cache')
        for component in components:
            pooldir = os.path.join('pool', component)
            if not os.path.isdir(pooldir):
                os.makedirs(pooldir)
            for arch in archs:
                if arch == 'source':
                    d = os.path.join('dists', pocket, component, 'source')
                else:
                    d = os.path.join('dists', pocket, component, 'binary-' + arch)
                if not os.path.isdir(d):
                    os.makedirs(d)
                if ref_archive:
                    logging.debug('checking refarchive (cur comp %s arch %s)', component, arch)
                    listfile = open(os.path.join('lists', '%s-%s' % (component, arch)), 'w')
                    for f in glob.glob(os.path.join(pooldir, '*/*/*%s%s' % (arch, ext))):
                        try:
                            (pkg, ver, suffix) = os.path.basename(f).split('_')
                        except ValueError:
                            logging.warning('Cannot split file name %s, unknown format; ignoring', f)
                            continue
                        # our package names can be "pkg" or "pkg-*" for pkg in the ref archive
                        refpkgs = [pkg]
                        try:
                            refpkgs.append(pkg[:pkg.rindex('-')])
                        except ValueError:
                            pass
                        for refpkg in refpkgs:
                            try:
                                if ref_archive[refpkg][pocket][ver][arch] == component:
                                    logging.debug('%s matches ref archive, adding', f)
                                    listfile.write(f)
                                    listfile.write('\n')
                                    break
                            except KeyError:
                                pass
                        else:
                            logging.debug('%s does not match ref archive', f)
                    listfile.close()
                else:
                    # just use everything we find
                    subprocess.call("find %s -name '*_%s%s' > lists/%s-%s" %
                                    (pooldir, arch, ext, component, arch), shell=True)

        # call apt-ftparchive
        subprocess.check_call(['apt-ftparchive', 'generate', conf.name], stderr=stderr)
        os.chdir(os.path.join('dists', pocket))
        apt_ftparchive = subprocess.Popen(
            ['apt-ftparchive',
             '-o', 'APT::FTPArchive::Release::Origin=Ubuntu',
             '-o', 'APT::FTPArchive::Release::Label=Ubuntu',
             '-o', 'APT::FTPArchive::Release::Suite=' + pocket,
             '-o', 'APT::FTPArchive::Release::Codename=' + pocket,
             '-o', 'APT::FTPArchive::Release::Architectures=' + ' '.join(sorted(archs)),
             '-o', 'APT::FTPArchive::Release::Components=' + ' '.join(sorted(components)),
             '-o', 'APT::FTPArchive::Release::Description=Debug symbols for Ubuntu ' + pocket,
             'release', '.'],
            stdout=subprocess.PIPE, stderr=stderr)
        out = apt_ftparchive.communicate()[0]
        assert apt_ftparchive.returncode == 0
        with open('Release', 'wb') as r:
            r.write(out)
        if gpg_sign:
            try:
                os.unlink('Release.gpg')
            except OSError:
                pass
            gpg = ['gpg', '-abs', '--batch',
                   '--personal-digest-preferences', 'SHA512,SHA256,SHA1']
            if gpg_keyid:
                gpg += ['--default-key', gpg_keyid]
            gpg += ['-o', 'Release.gpg', 'Release']
            assert subprocess.call(gpg, stdout=subprocess.PIPE, stderr=stderr) == 0
    finally:
        os.chdir(old_cwd)
        if os.path.isdir(os.path.join(archive_root, 'lists')):
            shutil.rmtree(os.path.join(archive_root, 'lists'))


def _packages_files(path, files):
    '''Add all files from given Packages path to the 'files' set.'''

    with open(path) as f:
        parser = apt_pkg.TagFile(f)
    while parser.step():
        f = parser.section['Filename']
        if f.startswith('./'):
            f = f[2:]
        files.add(f)
    parser.close()


def _sources_files(path, files):
    '''Add all files from given Sources path to the 'files' set.'''

    with open(path) as f:
        parser = apt_pkg.TagFile(f)
    while parser.step():
        d = parser.section.get('Directory')
        for l in parser.section['Files'].splitlines():
            fields = l.split()
            assert len(fields) == 3
            if d:
                files.add(os.path.join(d, fields[2]))
            else:
                files.add(fields[2])
    parser.close()


def archive_cleanup(archive_root, keep_days=0):
    '''Remove all files from an archive pool that are not mentioned in any
    Packages or Sources file and which are older than keep_days.
    '''
    old_cwd = os.getcwd()
    os.chdir(archive_root)

    try:
        # determine set of all indexed files (which we want to keep)
        distroot = 'dists'
        if not os.path.isdir(distroot):
            distroot = '.'

        indexed_files = set([])
        for root, dirs, files in os.walk(distroot):
            if root == '.':
                root = ''
            elif root.startswith('./'):
                root = root[2:]
            if 'Packages' in files:
                _packages_files(os.path.join(root, 'Packages'), indexed_files)
                indexed_files.add(os.path.join(root, 'Packages'))
                indexed_files.add(os.path.join(root, 'Packages.gz'))
                indexed_files.add(os.path.join(root, 'Packages.bz2'))
                indexed_files.add(os.path.join(root, 'Packages.xz'))
            else:
                assert 'Packages.gz' not in files
                assert 'Packages.bz2' not in files
                assert 'Packages.xz' not in files

            if 'Sources' in files:
                _sources_files(os.path.join(root, 'Sources'), indexed_files)
                indexed_files.add(os.path.join(root, 'Sources'))
                indexed_files.add(os.path.join(root, 'Sources.gz'))
                indexed_files.add(os.path.join(root, 'Sources.bz2'))
                indexed_files.add(os.path.join(root, 'Sources.xz'))
            else:
                assert 'Sources.gz' not in files
                assert 'Sources.bz2' not in files
                assert 'Sources.xz' not in files

        now = time.time()
        keep_seconds = keep_days * 24 * 3600

        # now walk the pool and clean up
        poolroot = 'pool'
        if not os.path.isdir(poolroot):
            poolroot = '.'
        for root, dirs, files in os.walk(poolroot):
            if root == '.':
                root = ''
            if root.startswith('./'):
                root = root[2:]
            for f in files:
                n = os.path.join(root, f)
                if n not in indexed_files:
                    if keep_days:
                        if (now - os.stat(n).st_mtime) <= keep_seconds:
                            continue
                    os.unlink(n)
    finally:
        os.chdir(old_cwd)


def _pack(src, dest):
    '''Pack uncompressed file src to compressed file dest.

    dest must end with .gz for gzip compression or .bz2 for bzip2
    compression. This uses a temporary file so that the resulting compressed
    file is not clobbered with a half-written new version.
    '''
    if dest.endswith('.gz'):
        z = gzip.open(dest + '.new', 'w', 9)
    elif dest.endswith('.bz2'):
        z = bz2.BZ2File(dest + '.new', 'w')
    f = open(src, 'rb')
    while True:
        block = f.read(10240)
        if not block:
            break
        z.write(block)
    f.close()
    z.close()
    os.rename(dest + '.new', dest)


def archive_dominate(archive_root, keep_days=0):
    '''Dominate all Sources and Packages files, refresh the corresponding *.gz
    files, and call archive_cleanup().

    keep_days is passed to archive_cleanup(), i. e. all files which are younger
    than keep_days will be retained.
    '''
    # determine set of all indexed files (which we want to keep)
    distroot = os.path.join(archive_root, 'dists')
    if not os.path.isdir(distroot):
        distroot = archive_root

    for root, dirs, files in os.walk(distroot):
        if 'Packages' in files:
            path = os.path.join(root, 'Packages')
            with open(path) as old, open(path + '.new', 'w') as new:
                debcontrol_dominate(old, new)

            # regenerate compressed files
            if 'Packages.gz' in files:
                _pack(path + '.new', path + '.gz')
            if 'Packages.bz2' in files:
                _pack(path + '.new', path + '.bz2')
            os.rename(path + '.new', path)
        else:
            assert 'Packages.gz' not in files
            assert 'Packages.bz2' not in files

        if 'Sources' in files:
            path = os.path.join(root, 'Sources')
            with open(path) as old, open(path + '.new', 'w') as new:
                debcontrol_dominate(old, new)

            # regenerate compressed files
            if 'Sources.gz' in files:
                _pack(path + '.new', path + '.gz')
            if 'Sources.bz2' in files:
                _pack(path + '.new', path + '.bz2')
            os.rename(path + '.new', path)
        else:
            assert 'Sources.gz' not in files
            assert 'Sources.bz2' not in files

    archive_cleanup(archive_root, keep_days)


class TestArchive:
    '''Test archive indexes'''

    def __init__(self):
        self.workdir = tempfile.mkdtemp()
        self.default_repo = os.path.join(self.workdir, 'ubuntu')
        self.default_repo_url = 'file://' + self.default_repo

    def __del__(self):
        shutil.rmtree(self.workdir)

    @classmethod
    def create_deb(kls, package, version, arch, dir, ext='.deb'):
        '''Create a dummy deb with given name, version, and architecture, and
        write it into the given directory.

        ext can specify a nonstandard file name suffix. Return the full path of
        the deb.'''

        d = tempfile.mkdtemp()
        os.mkdir(os.path.join(d, 'DEBIAN'))
        with open(os.path.join(d, 'DEBIAN', 'control'), 'w') as file:
            file.write('''Package: %s
Version: %s
Priority: optional
Section: devel
Architecture: %s
Maintainer: Joe <joe@joe.com>
Description: dummy package
 just a test dummy package
''' % (package, version, arch))

        if ':' in version:
            version = version.split(':', 1)[1]
        debpath = os.path.join(dir, '%s_%s_%s%s' % (package, version, arch, ext))
        subprocess.check_call(['dpkg', '-b', d, debpath],
                              stdout=subprocess.PIPE)

        shutil.rmtree(d)
        assert os.path.exists(debpath)
        return debpath

    def write_index(self, pocket, component, arches, packages, repo='ubuntu'):
        '''Write an index file into our test repo'''

        if not isinstance(arches, list):
            arches = [arches]

        for arch in arches:
            d = os.path.join(self.workdir, repo, 'dists', pocket, component, 'binary-' + arch)
            if not os.path.isdir(d):
                os.makedirs(d)
            with gzip.open(os.path.join(d, 'Packages.gz'), mode='wt') as f:
                for (name, version) in packages:
                    f.write('Package: %s\nVersion: %s\n\n' % (name, version))


#
# Test suite
#

import unittest
import sys


class _DebcontrolMapTest(unittest.TestCase):
    '''debcontrol_map()'''

    def setUp(self):

        self.tmpfiles = set(os.listdir('/tmp'))

        self.sources = tempfile.NamedTemporaryFile(mode='w')
        print('''Package: aalib
Binary: libaa-bin, libaa1-dev, libaa1
Version: 1.4p5-30
Directory: pool/main/a/aalib
Files:
 ad2e6b4821fb03fa6de7371a1b292f7b 733 aalib_1.4p5-30.dsc
 9801095c42bba12edebd1902bcf0a990 391028 aalib_1.4p5.orig.tar.gz

Package: abiword
Binary: abiword-plugins, abiword-plugins-gnome, abiword-help, abiword, abiword-common, abiword-gnome
Version: 2.4.4-0ubuntu5''', file=self.sources)
        self.sources.flush()

        self.packages = tempfile.NamedTemporaryFile(mode='w')
        print('''Package: abiword
Architecture: amd64
Version: 2.4.4-0ubuntu5
Description: Short
 Long
 Long

Package: abiword-common
Source: abiword
Version: 2.4.4-0ubuntu5''', file=self.packages)
        self.packages.flush()

    def tearDown(self):
        self.sources.close()
        self.packages.close()

        self.assertEqual(set(os.listdir('/tmp')), self.tmpfiles,
                         'does not leave temporary files around')

    def test_uncompressed(self):
        '''debcontrol_map() for uncompressed files'''

        i = debcontrol_map('file://' + self.sources.name)
        self.assertEqual(_mapkeystr(i), 'aalib abiword')
        self.assertEqual(_mapkeystr(i['aalib']), 'Binary Directory Files Version')
        self.assertEqual(_mapkeystr(i['abiword']), 'Binary Version')
        self.assertEqual(i['aalib']['Version'], '1.4p5-30')
        self.assertEqual(i['abiword']['Version'], '2.4.4-0ubuntu5')

        i = debcontrol_map('file://' + self.packages.name)
        self.assertEqual(_mapkeystr(i), 'abiword abiword-common')
        self.assertEqual(_mapkeystr(i['abiword']), 'Architecture Description Version')
        self.assertEqual(_mapkeystr(i['abiword-common']), 'Source Version')
        self.assertEqual(i['abiword']['Version'], '2.4.4-0ubuntu5')
        self.assertEqual(i['abiword']['Description'], 'Short\n Long\n Long')

        self.assertRaises(KeyError, i.__getitem__, 'nonexistant')
        self.assertRaises(KeyError, i['abiword'].__getitem__, 'nonexistant')

    def test_map_key(self):
        '''debcontrol_map for an alternate mapping key'''

        i = debcontrol_map('file://' + self.sources.name, 'Version')
        self.assertEqual(_mapkeystr(i), '1.4p5-30 2.4.4-0ubuntu5')
        self.assertEqual(_mapkeystr(i['1.4p5-30']), 'Binary Directory Files Package')

    def test_nonexisting(self):
        '''debcontrol_map() with a nonexisting URL'''

        self.assertRaises(IOError, debcontrol_map, 'file:///notexisting')


class _ArchiveMapTest(unittest.TestCase):
    '''archive_map()'''

    @classmethod
    def setUpClass(cls):
        cls.test_log = StringIO()
        logging.basicConfig(stream=cls.test_log)

    def setUp(self):
        self.tmpfiles = set(os.listdir('/tmp'))
        self.archive = TestArchive()

    def tearDown(self):
        del self.archive
        self.assertEqual(set(os.listdir('/tmp')), self.tmpfiles,
                         'does not leave temporary files around')

    def test_all_empty(self):
        self.assertEqual(archive_map({}, []), {})
        self.assertEqual(self.test_log.getvalue(), '')

    def test_empty_archive(self):
        self.assertEqual(archive_map({'testy': (['main'], ['vax'])},
                                     [self.archive.default_repo_url]),
                         {})
        self.assertIn('cannot find dists/testy/main/binary-vax/Packages',
                      self.test_log.getvalue())

    def test_with_sources(self):
        self.archive.write_index('testy', 'main', ['vax', '8088'],
                                 [('pi', '3.14'), ('libx1', '0.1-2')])
        self.archive.write_index('testy-updates', 'main', ['vax', '8088'],
                                 [('pi', '3.141'), ('libx1', '0.1-3'), ('e', '2.7')])
        self.archive.write_index('testy', 'extras', ['vax'], [('fancy', '2:1~1')])
        m = archive_map({'testy': (['main', 'extras'], ['vax', '8088'])},
                        [self.archive.default_repo_url])
        self.assertEqual(set(m.keys()), set(['libx1', 'pi', 'e', 'fancy']))
        self.assertEqual(m['libx1'],
                         {'testy': {'0.1-2': {'8088': 'main', 'vax': 'main'}},
                          'testy-updates': {'0.1-3': {'8088': 'main', 'vax': 'main'}}})
        self.assertEqual(m['fancy'],
                         {'testy': {'2:1~1': {'vax': 'extras'}}})

    def test_pockets_filter(self):
        self.archive.write_index('testy', 'main', ['vax', '8088'],
                                 [('pi', '3.14'), ('libx1', '0.1-2')])
        self.archive.write_index('testy-updates', 'main', ['vax', '8088'],
                                 [('pi', '3.141'), ('libx1', '0.1-3'), ('e', '2.7')])
        m = archive_map({'testy': (['main', 'extras'], ['vax', '8088'])},
                        [self.archive.default_repo_url],
                        pockets_filter=['testy-updates', 'somethingelse'])
        self.assertEqual(set(m.keys()), set(['libx1', 'pi', 'e']))
        self.assertEqual(m['libx1'],
                         {'testy-updates': {'0.1-3': {'8088': 'main', 'vax': 'main'}}})

    def test_mirror_fallback(self):
        self.archive.write_index('testy', 'main', 'vax', [('pkg', '1')])
        self.archive.write_index('testy-updates', 'extras', 'vax', [('pkg', '2')],
                                 repo='backports')
        m = archive_map({'testy': (['main', 'extras'], ['vax'])},
                        [self.archive.default_repo_url,
                         'file://%s/backports' % self.archive.workdir])
        self.assertEqual(m, {'pkg': {
            'testy': {'1': {'vax': 'main'}},
            'testy-updates': {'2': {'vax': 'extras'}}}})

    def test_strip_epoch(self):
        self.archive.write_index('testy', 'main', ['8088'],
                                 [('pi', '3.14'), ('fancy', '2:1~1')])
        m = archive_map({'testy': (['main'], ['8088'])},
                        [self.archive.default_repo_url], strip_epoch=True)
        self.assertEqual(m['pi'], {'testy': {'3.14': {'8088': 'main'}}})
        self.assertEqual(m['fancy'], {'testy': {'1~1': {'8088': 'main'}}})


def _ls(dir):
    '''Return an 'ls'-like string with a sorted space-separated list of a
    directory contents.'''

    l = os.listdir(dir)
    l.sort()
    return ' '.join(l)


def _mapkeystr(map):
    '''Return string with space-separated sorted list of keys of a
    dictionary.'''

    k = sorted(map)
    return ' '.join(k)


def _find_files(dir, strip_prefix=None):
    '''Return set of all files contained in given dir, and recursively, its
    subdirectories.

    If strip_prefix is given, it is removed from all file names.
    '''
    if strip_prefix:
        plen = len(strip_prefix)
    else:
        plen = 0
    result = set([])
    for root, dirs, files in os.walk(dir):
        for f in files:
            result.add(os.path.join(root, f)[plen:])
    return result


class _ArchiveTests(unittest.TestCase):
    '''Test functions that deal with an archive.'''

    def create_source(self, package, version, dir, native=False):
        '''Create a dummy source package with given name and version, and write
        it into the given directory.

        If 'native' is True, a native source package is created (dsc+tar.gz),
        otherwise a normal dsc/diff.gz/orig.tar.gz. Return a list of paths of
        all created files.'''

        dir = os.path.abspath(dir)
        d = tempfile.mkdtemp()
        old_cwd = os.getcwd()
        os.chdir(d)

        os.makedirs(os.path.join(package, 'debian'))
        with open(os.path.join(package, 'README'), 'w') as f:
            f.write('hello\n')
        with open(os.path.join(package, 'debian', 'control'), 'w') as f:
            f.write('''\
Source: %s
Maintainer: Joe <joe@ubuntu.com>

Package: %s
Architecture: all
''' % (package, package))
        with open(os.path.join(package, 'debian', 'changelog'), 'w') as f:
            f.write('''\
%s (%s) dummy; urgency=low

  * Dummy package.

 -- Joe <joe@joe.com>  Fri, 22 Sep 2006 13:07:40 +0200
''' % (package, version))
        with open(os.path.join(package, 'debian', 'rules'), 'w') as f:
            pass

        if not native:
            upstreamver = '-'.join(version.split('-')[:-1])
            t = tarfile.open('%s_%s.orig.tar.gz' % (package, upstreamver), 'w:gz')
            t.add(os.path.join(package, 'README'))
            t.close()

        subprocess.check_call(['dpkg-source', '-b', package],
                              stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        shutil.rmtree(package)
        created_files = []
        for f in os.listdir('.'):
            if os.path.exists(os.path.join(dir, f)):
                os.unlink(os.path.join(dir, f))
            shutil.move(f, dir)
            created_files.append(os.path.realpath(f))

        os.chdir(old_cwd)
        shutil.rmtree(d)
        return created_files

    def setUp(self):
        self.tmpfiles = set(os.listdir('/tmp'))

        self.archivedir = tempfile.mkdtemp('archivedir', 'archive_tools.py')
        self.testfiles_dir = tempfile.mkdtemp('testfiles', 'archive_tools.py')

        # create test tarballs
        d = tempfile.mkdtemp()
        old_cwd = os.getcwd()
        os.chdir(d)

        os.chdir(old_cwd)
        shutil.rmtree(d)

    def tearDown(self):
        shutil.rmtree(self.archivedir)
        shutil.rmtree(self.testfiles_dir)

        self.assertEqual(set(os.listdir('/tmp')), self.tmpfiles,
                         'does not leave temporary files around')

    def test_install_pool(self):
        '''install_pool()'''

        f = tempfile.NamedTemporaryFile(mode='w')
        print('hello', file=f)
        f.flush()

        self.assertTrue(install_pool(self.archivedir, f.name, 'main', 'src1'))
        self.assertTrue(install_pool(self.archivedir, f.name, 'main', 'src2'))
        self.assertTrue(install_pool(self.archivedir, f.name, 'main', 'libsrc'))
        self.assertTrue(install_pool(self.archivedir, f.name, 'universe', 'src'))
        self.assertFalse(install_pool(self.archivedir, f.name, 'universe', 'src'))

        self.assertEqual(_ls(os.path.join(self.archivedir, 'pool')), 'main universe')
        self.assertEqual(_ls(os.path.join(self.archivedir, 'pool', 'main')), 'libs s')
        self.assertEqual(_ls(os.path.join(self.archivedir, 'pool', 'main', 's')), 'src1 src2')
        self.assertEqual(_ls(os.path.join(self.archivedir, 'pool', 'main', 'libs')), 'libsrc')
        self.assertEqual(_ls(os.path.join(self.archivedir, 'pool', 'main', 'libs', 'libsrc')),
                         os.path.basename(f.name))
        with open(os.path.join(self.archivedir, 'pool', 'main', 'libs',
                               'libsrc', os.path.basename(f.name))) as f:
            self.assertEqual(f.read(), 'hello\n')

        f.close()

    def test_create_indexes(self):
        '''create_indexes()'''

        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'foo1', '1-1', 'i386', self.testfiles_dir), 'universe', 'foo'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'foo2', '1-1', 'i386', self.testfiles_dir), 'universe', 'foo'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'bar1', '2-2ubuntu1', 'powerpc', self.testfiles_dir), 'main', 'bar'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'bar2', '2-2ubuntu1', 'powerpc', self.testfiles_dir), 'main', 'bar'))
        create_indexes(self.archivedir, 'edgy', ['main', 'restricted', 'universe'],
                       ['i386', 'powerpc', 'amd64'], stderr=subprocess.PIPE)

        # check directory structure
        self.assertEqual(_ls(self.archivedir), '.cache dists pool')
        self.assertTrue(os.path.exists(os.path.join(self.archivedir, '.cache', 'db')))
        self.assertEqual(_ls(os.path.join(self.archivedir, 'dists', 'edgy')),
                         'Release main restricted universe')
        self.assertEqual(_ls(os.path.join(self.archivedir, 'dists', 'edgy', 'main', 'binary-i386')),
                         'Packages Packages.gz Packages.xz')
        self.assertEqual(_ls(os.path.join(self.archivedir, 'dists', 'edgy', 'universe', 'binary-powerpc')),
                         'Packages Packages.gz Packages.xz')

        # check Packages contents
        map = debcontrol_map(os.path.join(
            self.archivedir, 'dists', 'edgy', 'universe', 'binary-i386', 'Packages.gz'))
        self.assertEqual(map['foo1']['Architecture'], 'i386')
        self.assertEqual(map['foo1']['Version'], '1-1')
        self.assertEqual(map['foo1']['Filename'], 'pool/universe/f/foo/foo1_1-1_i386.deb')
        self.assertNotIn('bar1', map)  # since that's powerpc

        map = debcontrol_map(os.path.join(
            self.archivedir, 'dists', 'edgy', 'main', 'binary-powerpc', 'Packages.gz'))
        self.assertEqual(map['bar2']['Architecture'], 'powerpc')
        self.assertEqual(map['bar2']['Version'], '2-2ubuntu1')
        self.assertEqual(map['bar2']['Filename'], 'pool/main/b/bar/bar2_2-2ubuntu1_powerpc.deb')
        self.assertNotIn('foo1', map)  # since that's i386

        xzmap = debcontrol_map(os.path.join(
            self.archivedir, 'dists', 'edgy', 'main', 'binary-powerpc', 'Packages.xz'))
        self.assertEqual(map, xzmap)

        with open(os.path.join(self.archivedir, 'dists', 'edgy', 'main',
                               'binary-i386', 'Packages')) as f:
            self.assertEqual(f.read(), '')
        with open(os.path.join(self.archivedir, 'dists', 'edgy', 'universe',
                               'binary-powerpc', 'Packages')) as f:
            self.assertEqual(f.read(), '')
        with open(os.path.join(self.archivedir, 'dists', 'edgy', 'main',
                               'binary-amd64', 'Packages')) as f:
            self.assertEqual(f.read(), '')
        with open(os.path.join(self.archivedir, 'dists', 'edgy', 'universe',
                               'binary-amd64', 'Packages')) as f:
            self.assertEqual(f.read(), '')

        with open(os.path.join(self.archivedir, 'dists', 'edgy', 'Release')) as f:
            rel = f.read()
        self.assertIn('\nSuite: edgy\n', rel)
        self.assertIn('Architectures: amd64 i386 powerpc\n', rel)
        self.assertIn('\nComponents: main restricted universe\n', rel)
        self.assertIn('\nOrigin: Ubuntu', rel)
        self.assertIn('\nLabel: Ubuntu', rel)
        self.assertIn('\nDescription: ', rel)
        self.assertNotIn('dists/edgy', rel)
        self.assertNotIn('Release', rel)
        self.assertIn(' main/binary-amd64/Packages\n', rel)

    def test_create_indexes_source(self):
        '''create_indexes() for source packages'''

        maindir = os.path.join(self.archivedir, 'pool', 'main')
        os.makedirs(os.path.join(maindir, 'f', 'foo'))
        self.create_source('foo', '1-1', os.path.join(maindir, 'f', 'foo'))
        os.makedirs(os.path.join(maindir, 'b', 'bar'))
        self.create_source('bar', '2-2', os.path.join(maindir, 'b', 'bar'), native=True)

        create_indexes(self.archivedir, 'edgy', ['main'], ['source', 'i386'],
                       stderr=subprocess.PIPE)

        # check directory structure
        self.assertEqual(_ls(self.archivedir), '.cache dists pool')
        self.assertEqual(_ls(os.path.join(self.archivedir, 'dists', 'edgy')),
                         'Release main')
        self.assertEqual(_ls(os.path.join(self.archivedir, 'dists', 'edgy', 'main')),
                         'binary-i386 source')
        self.assertEqual(_ls(os.path.join(
            self.archivedir, 'dists', 'edgy', 'main', 'binary-i386')),
            'Packages Packages.gz Packages.xz')
        self.assertEqual(_ls(os.path.join(
            self.archivedir, 'dists', 'edgy', 'main', 'source')),
            'Sources Sources.gz Sources.xz')
        with open(os.path.join(self.archivedir, 'dists', 'edgy', 'main',
                               'binary-i386', 'Packages')) as f:
            self.assertEqual(f.read(), '')
        map = debcontrol_map(os.path.join(
            self.archivedir, 'dists', 'edgy', 'main', 'source', 'Sources'))

        # check generated Sources validity
        self.assertEqual(_mapkeystr(map), 'bar foo')
        self.assertEqual(map['bar']['Version'], '2-2')
        self.assertEqual(map['bar']['Directory'], 'pool/main/b/bar')
        self.assertGreater(map['bar']['Files'].find('bar_2-2.tar.gz'), 0)
        self.assertGreater(map['bar']['Files'].find('bar_2-2.dsc'), 0)
        self.assertEqual(map['foo']['Version'], '1-1')
        self.assertEqual(map['foo']['Directory'], 'pool/main/f/foo')
        self.assertGreater(map['foo']['Files'].find('foo_1.orig.tar.gz'), 0)
        self.assertGreater(map['foo']['Files'].find('foo_1-1.diff.gz'), 0)
        self.assertGreater(map['foo']['Files'].find('foo_1-1.dsc'), 0)

        xzmap = debcontrol_map(os.path.join(
            self.archivedir, 'dists', 'edgy', 'main', 'source', 'Sources.xz'))
        self.assertEqual(map, xzmap)

    def test_create_indexes_altext(self):
        '''create_indexes() with a non-default file name extension.'''

        # install some test debs
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'foo1', '1-1', 'i386', self.testfiles_dir, '.ddeb'),
            'universe', 'foo'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'foo2', '1-1', 'i386', self.testfiles_dir, '.ddeb'),
            'universe', 'foo'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'bar1', '2-2ubuntu1', 'powerpc', self.testfiles_dir, '.ddeb'),
            'main', 'bar'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'bar2', '2-2ubuntu1', 'powerpc', self.testfiles_dir, '.ddeb'),
            'main', 'bar'))

        # call create_indexes()
        create_indexes(self.archivedir, 'edgy', ['main', 'restricted', 'universe'],
                       ['i386', 'powerpc', 'amd64'], ext='.ddeb', stderr=subprocess.PIPE)

        # check directory structure
        self.assertEqual(_ls(self.archivedir), '.cache dists pool')
        self.assertEqual(_ls(os.path.join(
            self.archivedir, 'dists', 'edgy', 'main', 'binary-i386')),
            'Packages Packages.gz Packages.xz')
        self.assertEqual(_ls(os.path.join(
            self.archivedir, 'dists', 'edgy', 'universe', 'binary-powerpc')),
            'Packages Packages.gz Packages.xz')

        # check Packages contents
        map = debcontrol_map(os.path.join(
            self.archivedir, 'dists', 'edgy', 'universe', 'binary-i386', 'Packages.gz'))
        self.assertEqual(map['foo1']['Architecture'], 'i386')
        self.assertEqual(map['foo1']['Version'], '1-1')
        self.assertEqual(map['foo1']['Filename'], 'pool/universe/f/foo/foo1_1-1_i386.ddeb')
        self.assertNotIn('bar1', map)  # since that's powerpc

        xzmap = debcontrol_map(os.path.join(
            self.archivedir, 'dists', 'edgy', 'universe', 'binary-i386', 'Packages.xz'))
        self.assertEqual(map, xzmap)

        with open(os.path.join(self.archivedir, 'dists', 'edgy',
                               'main', 'binary-i386', 'Packages')) as f:
            self.assertEqual(f.read(),'')

        self.assertGreater(os.path.getsize(
            os.path.join(self.archivedir, 'dists', 'edgy', 'Release')), 0)

    def test_create_indexes_performance(self):
        '''performance of create_indexes()'''

        # install some test debs
        for i in range(1000):
            install_pool(self.archivedir, TestArchive.create_deb(
                str(i), '1-1', 'i386', self.testfiles_dir), 'main', str(i))

        print('[setup done]')
        sys.stdout.flush()

        # call create_indexes()
        t1 = time.time()
        create_indexes(self.archivedir, 'edgy', ['main'], ['i386'], stderr=subprocess.PIPE)
        t2 = time.time()

        pkg = os.path.join(self.archivedir, 'dists', 'edgy', 'main', 'binary-i386', 'Packages')
        print('Packages:', os.path.getsize(pkg), 'Packages.gz:', os.path.getsize(pkg + '.gz'))
        print('%.1f seconds ' % (t2 - t1))
        sys.stdout.flush()

    def test_create_indexes_refarchive(self):
        '''create_indexes() with a reference archive'''

        # install some test debs

        # ... crufty old version
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'foo-dbgsym', '1-0', 'i386', self.testfiles_dir, '.ddeb'),
            'universe', 'foo'))

        # ... dapper
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'foo-dbgsym', '1-1', 'i386', self.testfiles_dir, '.ddeb'),
            'universe', 'foo'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'bar1-dbgsym', '2-2', 'powerpc', self.testfiles_dir, '.ddeb'),
            'main', 'bar'))

        # ... edgy
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'foo-dbgsym', '1-2', 'i386', self.testfiles_dir, '.ddeb'),
            'universe', 'foo'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'bar1-dbgsym', '2-2ubuntu1', 'powerpc', self.testfiles_dir, '.ddeb'),
            'main', 'bar'))

        ref_archive = tempfile.mkdtemp()
        try:
            # create reference indexes
            indexes = {
                'dapper': {
                    'main': {'powerpc': 'Package: bar1\nSource:bar\nVersion: 2-2\n'},
                    'universe': {'i386': 'Package: foo\nVersion: 1-1\n'}
                },
                'edgy': {
                    'main': {'powerpc': 'Package: bar1\nSource:bar\nVersion: 2-2ubuntu1\n'},
                    'universe': {'i386': 'Package: foo\nVersion: 1-2\n'}
                }
            }
            for release in ('dapper', 'edgy'):
                for comp in ('main', 'universe'):
                    for arch in ('i386', 'powerpc'):
                        dir = os.path.join(ref_archive, 'dists', release,
                                           comp, 'binary-' + arch)
                        os.makedirs(dir)
                        p = os.path.join(dir, 'Packages')
                        f = open(p, 'w')
                        try:
                            f.write(indexes[release][comp][arch])
                        except KeyError:
                            pass
                        f.close()
                        _pack(p, p + '.gz')

            comp_arch = (['main', 'universe'], ['i386', 'powerpc'])
            amap = archive_map({'dapper': comp_arch, 'edgy': comp_arch},
                               ['file://' + ref_archive],
                               pockets_filter=['dapper', 'edgy'],
                               strip_epoch=True)

            # create our indexes
            create_indexes(self.archivedir, 'dapper', ['main', 'universe'],
                           ['i386', 'powerpc'], stderr=subprocess.PIPE, ext='.ddeb',
                           ref_archive=amap)
            create_indexes(self.archivedir, 'edgy', ['main', 'universe'],
                           ['i386', 'powerpc'], stderr=subprocess.PIPE, ext='.ddeb',
                           ref_archive=amap)

            # check correct versions in Packages files

            map = debcontrol_map(os.path.join(
                self.archivedir, 'dists', 'dapper', 'universe', 'binary-i386', 'Packages.gz'),
                'Version')
            self.assertEqual(list(map.keys()), ['1-1'])
            self.assertEqual(map['1-1']['Package'], 'foo-dbgsym')

            map = debcontrol_map(os.path.join(
                self.archivedir, 'dists', 'dapper', 'main', 'binary-powerpc', 'Packages.gz'),
                'Version')
            self.assertEqual(list(map.keys()), ['2-2'])
            self.assertEqual(map['2-2']['Package'], 'bar1-dbgsym')

            map = debcontrol_map(os.path.join(
                self.archivedir, 'dists', 'edgy', 'universe', 'binary-i386', 'Packages.gz'),
                'Version')
            self.assertEqual(list(map.keys()), ['1-2'])
            self.assertEqual(map['1-2']['Package'], 'foo-dbgsym')

            map = debcontrol_map(os.path.join(
                self.archivedir, 'dists', 'edgy', 'main', 'binary-powerpc', 'Packages.gz'),
                'Version')
            self.assertEqual(list(map.keys()), ['2-2ubuntu1'])
            self.assertEqual(map['2-2ubuntu1']['Package'], 'bar1-dbgsym')

        finally:
            shutil.rmtree(ref_archive)

    def test_debcontrol_dominate(self):
        '''debcontrol_dominate()'''

        i = tempfile.NamedTemporaryFile(mode='w')
        i.write('''Package: foo
Version: 1-0
Description: oldest foo

Package: bar
Version: 1-0
Description: barbar

Package: foo
Version: 1:1-10
Description: newest foo
 I wanna have this!

Package: foo
Version: 1:1-5
Description: slightly older foo
 (needs proper VersionCompare)

Package: foo
Version: 3-1
Description: old foo crap
''')
        i.flush()

        o = StringIO()
        debcontrol_dominate(open(i.name), o)

        self.assertEqual(o.getvalue(), '''Package: bar
Version: 1-0
Description: barbar

Package: foo
Version: 1:1-10
Description: newest foo
 I wanna have this!
''')

    def test_archive_cleanup(self):
        '''archive_cleanup()'''

        # install some test debs
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'foo1', '1-1', 'i386', self.testfiles_dir), 'universe', 'foo'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'foo2', '1-1', 'i386', self.testfiles_dir), 'universe', 'foo'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'bar1', '2-2ubuntu1', 'powerpc', self.testfiles_dir), 'main', 'bar'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'bar2', '2-2ubuntu1', 'powerpc', self.testfiles_dir), 'main', 'bar'))

        # call create_indexes()
        create_indexes(self.archivedir, 'edgy', ['main', 'restricted', 'universe'],
                       ['i386', 'powerpc', 'amd64'], stderr=subprocess.PIPE)

        # make file system snapshot, clean up, check that nothing was removed
        clean_file_list = _find_files(self.archivedir)
        archive_cleanup(self.archivedir)
        self.assertEqual(_find_files(self.archivedir), clean_file_list,
                         'missing: ' + str(clean_file_list.difference(_find_files(self.archivedir))))

        # now add some cruft
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'foo1', '1-0', 'i386', self.testfiles_dir), 'universe', 'foo'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'bar1', '2-2', 'powerpc', self.testfiles_dir), 'main', 'bar'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'cruft', '5', 'amd64', self.testfiles_dir), 'main', 'cruft'))

        # test with keeping old files around; make cruft a recent and bar a > 2
        # days old deb
        bar_path = os.path.join(self.archivedir, 'pool', 'main', 'b', 'bar',
                                'bar1_2-2_powerpc.deb')
        os.utime(bar_path, (time.time(), time.time() - 2.1 * 24 * 3600))

        archive_cleanup(self.archivedir, keep_days=2)
        self.assertTrue(os.path.exists(os.path.join(
            self.archivedir, 'pool', 'main', 'c', 'cruft', 'cruft_5_amd64.deb')))
        self.assertFalse(os.path.exists(bar_path))

        # now clean up everything
        archive_cleanup(self.archivedir)
        self.assertEqual(_find_files(self.archivedir), clean_file_list)

    def test_archive_cleanup_customlayout(self):
        '''archive_cleanup() with a non-standard archive structure'''

        with open(os.path.join(self.archivedir, 'foo_1_i386.deb'), 'w') as f:
            f.write('foo_1')
        with open(os.path.join(self.archivedir, 'foo_2_i386.deb'), 'w') as f:
            f.write('foo_2')
        with open(os.path.join(self.archivedir, 'bar_1_i386.deb'), 'w') as f:
            f.write('bar_1')
        with open(os.path.join(self.archivedir, 'cruft_3_i386.deb'), 'w') as f:
            f.write('cruft')
        with open(os.path.join(self.archivedir, 'Packages'), 'w') as f:
            f.write('''Package: foo
Version: 2
Filename: foo_2_i386.deb

Package: bar
Version: 1
Filename: ./bar_1_i386.deb
''')
        archive_cleanup(self.archivedir)
        self.assertEqual(_find_files(self.archivedir), set([
            os.path.join(self.archivedir, 'foo_2_i386.deb'),
            os.path.join(self.archivedir, 'bar_1_i386.deb'),
            os.path.join(self.archivedir, 'Packages')]))

    def test_archive_dominate(self):
        '''archive_dominate()'''

        # install some test debs and call these 'dapper'
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'foo', '1-0', 'i386', self.testfiles_dir), 'universe', 'foo'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'foo', '1-1', 'i386', self.testfiles_dir), 'universe', 'foo'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'bar', '2-2', 'i386', self.testfiles_dir), 'main', 'bar'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'bar', '2-2ubuntu1', 'i386', self.testfiles_dir), 'main', 'bar'))
        self.create_source('foo', '1-0', os.path.join(
            self.archivedir, 'pool', 'universe', 'f', 'foo'))
        self.create_source('foo', '1-1', os.path.join(
            self.archivedir, 'pool', 'universe', 'f', 'foo'))
        self.create_source('bar', '2-2', os.path.join(
            self.archivedir, 'pool', 'main', 'b', 'bar'))
        self.create_source('bar', '2-2ubuntu1', os.path.join(
            self.archivedir, 'pool', 'main', 'b', 'bar'), native=True)

        create_indexes(self.archivedir, 'dapper', ['main', 'universe'],
                       ['source', 'i386'], stderr=subprocess.PIPE)

        # install some newer versions, call these 'edgy'
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'foo', '1-2', 'i386', self.testfiles_dir), 'universe', 'foo'))
        self.assertTrue(install_pool(self.archivedir, TestArchive.create_deb(
            'bar', '3-0ubuntu1', 'i386', self.testfiles_dir), 'main', 'bar'))
        self.create_source('foo', '1-2', os.path.join(self.archivedir, 'pool',
                                                      'universe', 'f', 'foo'))
        self.create_source('bar', '3-0ubuntu1', os.path.join(
            self.archivedir, 'pool', 'main', 'b', 'bar'), native=True)

        create_indexes(self.archivedir, 'edgy', ['main', 'universe'],
                       ['source', 'i386'], stderr=subprocess.PIPE)

        archive_dominate(self.archivedir)

        # check that we only have the latest stuff from dapper and edgy
        self.assertEqual(
            _find_files(os.path.join(self.archivedir, 'pool'),
                        strip_prefix=self.archivedir),
            set(['/pool/main/b/bar/bar_2-2ubuntu1.dsc',
                 '/pool/main/b/bar/bar_2-2ubuntu1.tar.gz',
                 '/pool/main/b/bar/bar_2-2ubuntu1_i386.deb',
                 '/pool/main/b/bar/bar_3-0ubuntu1.dsc',
                 '/pool/main/b/bar/bar_3-0ubuntu1.tar.gz',
                 '/pool/main/b/bar/bar_3-0ubuntu1_i386.deb',
                 '/pool/universe/f/foo/foo_1-1.diff.gz',
                 '/pool/universe/f/foo/foo_1-1.dsc',
                 '/pool/universe/f/foo/foo_1-1_i386.deb',
                 '/pool/universe/f/foo/foo_1-2.diff.gz',
                 '/pool/universe/f/foo/foo_1-2.dsc',
                 '/pool/universe/f/foo/foo_1-2_i386.deb',
                 '/pool/universe/f/foo/foo_1.orig.tar.gz']))

        # check that Packages.gz and Sources.gz are re-created properly
        s = os.path.join(self.archivedir, 'dists', 'dapper', 'main',
                         'source', 'Sources')
        with open(s, mode='rb') as f:
            self.assertEqual(f.read(), gzip.open(s + '.gz').read())
        s = os.path.join(self.archivedir, 'dists', 'dapper', 'main',
                         'binary-i386', 'Packages')
        with open(s, mode='rb') as f:
            self.assertEqual(f.read(), gzip.open(s + '.gz').read())
        s = os.path.join(self.archivedir, 'dists', 'edgy', 'universe',
                         'source', 'Sources')
        with open(s, mode='rb') as f:
            self.assertEqual(f.read(), gzip.open(s + '.gz').read())
        s = os.path.join(self.archivedir, 'dists', 'edgy', 'universe',
                         'binary-i386', 'Packages')
        with open(s, mode='rb') as f:
            self.assertEqual(f.read(), gzip.open(s + '.gz').read())

    def test_archive_dominate_flat(self):
        '''archive_dominate() with a flat tree'''

        # install some test debs and call these 'dapper'
        self.assertTrue(TestArchive.create_deb('foo', '1-0', 'i386', self.archivedir))
        self.assertTrue(TestArchive.create_deb('foo', '1-1', 'i386', self.archivedir))
        self.create_source('foo', '1-0', self.archivedir)
        self.create_source('foo', '1-1', self.archivedir)

        # create indices
        old_cwd = os.getcwd()
        try:
            os.chdir(self.archivedir)
            subprocess.check_call('apt-ftparchive packages . > Packages',
                                  shell=True)
            subprocess.check_call('apt-ftparchive sources . > Sources',
                                  shell=True, stderr=subprocess.PIPE)
        finally:
            os.chdir(old_cwd)

        archive_dominate(self.archivedir)

        self.assertEqual(_find_files(self.archivedir, strip_prefix=self.archivedir),
                         set(['/Packages', '/Sources', '/foo_1-1_i386.deb',
                              '/foo_1.orig.tar.gz', '/foo_1-1.dsc',
                              '/foo_1-1.diff.gz']))


if __name__ == '__main__':
    unittest.main()
