#!/usr/bin/python3
import os.path
import logging
import urllib
import argparse
import fcntl
import datetime

import wadllib.iso_strptime

import lpinfo
import archive_tools


def parse_args(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument('-r', '--archive-root',
                        default=os.path.expanduser('~/www'),
                        help='ddeb archive root directory (default: %(default)s)')
    parser.add_argument('-d', '--download-only', default=False,
                        action='store_true',
                        help='Only download ddebs, do not update indexes')
    parser.add_argument('-v', '--verbose', default=False,
                        action='store_true',
                        help='Show debug messages')
    parser.add_argument('-q', '--quiet', default=False,
                        action='store_true',
                        help='Only show warnings and errors')
    parser.add_argument('-g', '--no-gpg-sign', default=False,
                        action='store_true',
                        help='Disable GPG signing of Release files')
    parser.add_argument('-c', '--no-clean', default=False,
                        action='store_true',
                        help='Disable archive cleanup')
    parser.add_argument('--regenerate-all', default=False,
                        action='store_true',
                        help='Force index regeneration for all series/pockets')
    parser.add_argument('--distribution', default='ubuntu',
                        help='Launchpad distribution name (default: %(default)s)')
    parser.add_argument('--mirror', action='append',
                        help='Distro archive mirrors, can be specified multiple '
                        'times; defaults to default distro archive')

    return parser.parse_args(args)


distroarchseries_links = {}
distroseries_links = {}


def get_suite_for_publication(pub):
    '''Return the suite for a binary_package_publishing_history object.

    We memoise the series objects, as otherwise this is very slow for lots
    of publications.
    '''
    # Don't be tempted to shorten this using setdefault; fetching
    # pub.distro_arch_series and das.distroseries is expensive.
    distroarchseries_link = pub.distro_arch_series_link
    if distroarchseries_link not in distroarchseries_links:
        distroarchseries_links[distroarchseries_link] = pub.distro_arch_series
    distroarchseries = distroarchseries_links[distroarchseries_link]
    distroseries_link = distroarchseries.distroseries_link
    if distroseries_link not in distroseries_links:
        distroseries_links[distroseries_link] = distroarchseries.distroseries
    distroseries = distroseries_links[distroseries_link]
    return lpinfo.make_suite(distroseries, pub.pocket)


def install_from_librarian(pub, url, ddeb_archive_root):
    '''Fetch ddeb from the librarian and install it into ddeb_archive_root

    If the file already exists in the appropriate place in pool, it will not
    be downloaded again.

    Return True if it was installed successfully or already present,
    otherwise False.
    '''
    ddeb = urllib.parse.unquote(os.path.basename(url))
    try:
        (dbgsymname, version, arch) = ddeb.split('_')
        assert arch.endswith('.ddeb')
        (arch, _) = arch.split('.')
        assert dbgsymname.endswith('-dbgsym')
    except (ValueError, AssertionError):
        logging.warning('Ignoring non-ddeb file %s in %s', ddeb, pub)
        return False

    try:
        source_name = pub.build.source_package_name
    except Exception:
        logging.exception('Failed to get source package name for %s', pub)
        raise
    if source_name.startswith('lib'):
        prefix = source_name[:4]
    else:
        prefix = source_name[0]

    destdir = os.path.join(
        ddeb_archive_root, 'pool', pub.component_name, prefix, source_name)
    dest = os.path.join(destdir, ddeb)
    if os.path.exists(dest):
        logging.debug('%s already exists, skipping', dest)
    else:
        logging.debug('Downloading %s into %s', ddeb, os.path.dirname(dest))
        try:
            os.makedirs(destdir)
        except OSError:
            pass
        with urllib.request.urlopen(url) as req:
            if req.getcode() != 200:
                logging.error('URL %s failed with code %u', req.geturl(), code)
                return False
            response = req.read()
        with open(dest, 'wb') as handle:
            handle.write(response)

        logging.debug('Downloaded %s into %s', ddeb, os.path.dirname(dest))

    return True


def main(argv=None):
    args = parse_args(argv)

    if args.verbose:
        level = logging.DEBUG
    elif args.quiet:
        level = logging.WARNING
    else:
        level = logging.INFO
    logging.basicConfig(level=level, format='%(levelname)s: %(message)s')

    logging.debug('arguments: %s', args)

    # avoid multiple ddeb-retrievers stomping on each other
    if not os.path.isdir(args.archive_root):
        os.makedirs(args.archive_root)
    lockfd = os.open(os.path.join(args.archive_root, '.lock'), os.O_CREAT | os.O_WRONLY)
    try:
        fcntl.flock(lockfd, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except (IOError, OSError):
        logging.error('Cannot acquire lock, another ddeb-retriever instance is already running')
        return

    UTC = wadllib.iso_strptime.TimeZone('+00:00')
    lp_threshold_path = os.path.join(args.archive_root, '.lp-threshold')
    try:
        with open(lp_threshold_path) as lp_threshold_file:
            lp_threshold = datetime.datetime.fromtimestamp(
                int(lp_threshold_file.readline()), tz=UTC)
    except IOError:
        lp_threshold = None

    series_info = lpinfo.get_series(args.distribution)
    logging.debug('series info: %s', series_info)

    if lp_threshold is not None:
        # Allow a grace period to cope with publications arriving out of
        # order during long transactions.
        real_threshold = lp_threshold - datetime.timedelta(hours=1)
        logging.info(
            'Retrieving Launchpad publications since %s' % real_threshold)
    else:
        real_threshold = None
        logging.info('Retrieving all Launchpad publications')
    binary_pubs = lpinfo.get_binary_publications(
        args.distribution, created_since_date=real_threshold)
    debug_pubs = []
    latest_date_created = lp_threshold
    for pub in binary_pubs:
        if pub.status not in ('Published', 'Pending'):
            continue
        if (latest_date_created is None or
                pub.date_created > latest_date_created):
            latest_date_created = pub.date_created
        if pub.is_debug:
            debug_pubs.append(pub)
            logging.debug('Got ddeb: %s (pocket: %s)', pub.display_name, pub.pocket)

    # build map of package versions/components in affected pockets
    if args.regenerate_all:
        updated_pockets = set(series_info.keys())
        logging.info('Regenerating all indexes for series: %s', ' '.join(updated_pockets))
    else:
        updated_pockets = set()
        for pub in debug_pubs:
            updated_pockets.add(get_suite_for_publication(pub))
        logging.info('got downloaded/queued ddebs for pockets: %s', ' '.join(updated_pockets))

    # complete the pocket list
    for p in list(updated_pockets):
        # for transient pockets like -proposed we need to include the release
        # pocket
        if '-' in p:
            rel = p.rsplit('-', 1)[0]
            updated_pockets.add(rel)
        else:
            # for release pockets we need to include the SRU pockets
            updated_pockets.add(p + '-proposed')
            updated_pockets.add(p + '-updates')
    logging.debug('release-extended pockets: %s', ' '.join(updated_pockets))

    logging.info('Building archive map, this can take a while')
    archive_map = archive_tools.archive_map(series_info, args.mirror or None,
                                            pockets_filter=updated_pockets,
                                            strip_epoch=True)

    logging.info('Building ddeb map, this can take a while')
    ddeb_map = archive_tools.archive_map(series_info, [args.archive_root],
                                         ignore_missing=True)

    # try to install ddebs from publishing history records
    installed_pockets = set()
    for pub in debug_pubs:
        already_present = False
        pub_installed = False
        if pub.binary_package_name in ddeb_map:
            existing = ddeb_map[pub.binary_package_name]
            for _, series_map in existing.items():
                if pub.distro_arch_series.architecture_tag in series_map.get(pub.binary_package_version, {}):
                    already_present = True
                    break
        if already_present:
            # We already have this ddeb on disk, so just note that indexes
            # for this pocket need to be rebuilt.
            pub_installed = True
        else:
            for url in pub.binaryFileUrls():
                if install_from_librarian(pub, url, args.archive_root):
                    pub_installed = True
        if pub_installed:
            installed_pockets.add(get_suite_for_publication(pub))

    if latest_date_created is not None:
        epoch = datetime.datetime.fromtimestamp(0, tz=UTC)
        new_threshold = (latest_date_created - epoch).total_seconds()
        with open(lp_threshold_path, 'w') as lp_threshold_file:
            lp_threshold_file.write("%d\n" % new_threshold)

    if args.download_only:
        os.close(lockfd)
        return

    logging.debug('pockets that got installed to: %s', installed_pockets)

    # build indexes
    for series, sinfo in series_info.items():
        for pocket in ['', '-updates', '-proposed']:
            if not args.regenerate_all and series + pocket not in installed_pockets:
                continue
            logging.info('Building indexes for %s%s', series, pocket)
            archive_tools.create_indexes(args.archive_root, series + pocket,
                                         sinfo[0], sinfo[1],
                                         ext='.ddeb',
                                         ref_archive=archive_map,
                                         gpg_sign=not args.no_gpg_sign)

    if not args.no_clean and installed_pockets:
        logging.info('Cleaning up archive...')
        archive_tools.archive_cleanup(args.archive_root, 30)

    os.close(lockfd)
